@@ -13,15 +13,28 @@
 """Discover and lookup command plugins.
 """
 
-import inspect
 import logging
 
 import pkg_resources
 
+from . import utils
+
 
 LOG = logging.getLogger(__name__)
 
 
+def _get_commands_by_partial_name(args, commands):
+    n = len(args)
+    candidates = []
+    for command_name in commands:
+        command_parts = command_name.split()
+        if len(command_parts) != n:
+            continue
+        if all(command_parts[i].startswith(args[i]) for i in range(n)):
+            candidates.append(command_name)
+    return candidates
+
+
 class EntryPointWrapper(object):
     """Wrap up a command class already imported to make it look like a plugin.
     """
@@ -45,16 +58,20 @@ class CommandManager(object):
     """
     def __init__(self, namespace, convert_underscores=True):
         self.commands = {}
+        self._legacy = {}
         self.namespace = namespace
         self.convert_underscores = convert_underscores
+        self.group_list = []
         self._load_commands()
 
     def _load_commands(self):
         # NOTE(jamielennox): kept for compatibility.
-        self.load_commands(self.namespace)
+        if self.namespace:
+            self.load_commands(self.namespace)
 
     def load_commands(self, namespace):
         """Load all the commands from an entrypoint"""
+        self.group_list.append(namespace)
         for ep in pkg_resources.iter_entry_points(namespace):
             LOG.debug('found command %r', ep.name)
             cmd_name = (ep.name.replace('_', ' ')
@@ -69,31 +86,85 @@ class CommandManager(object):
     def add_command(self, name, command_class):
         self.commands[name] = EntryPointWrapper(name, command_class)
 
+    def add_legacy_command(self, old_name, new_name):
+        """Map an old command name to the new name.
+
+        :param old_name: The old command name.
+        :type old_name: str
+        :param new_name: The new command name.
+        :type new_name: str
+
+        """
+        self._legacy[old_name] = new_name
+
     def find_command(self, argv):
         """Given an argument list, find a command and
         return the processor and any remaining arguments.
         """
-        search_args = argv[:]
-        name = ''
-        while search_args:
-            if search_args[0].startswith('-'):
-                name = '%s %s' % (name, search_args[0])
-                raise ValueError('Invalid command %r' % name)
-            next_val = search_args.pop(0)
-            name = '%s %s' % (name, next_val) if name else next_val
+        start = self._get_last_possible_command_index(argv)
+        for i in range(start, 0, -1):
+            name = ' '.join(argv[:i])
+            search_args = argv[i:]
+            # The legacy command handling may modify name, so remember
+            # the value we actually found in argv so we can return it.
+            return_name = name
+            # Convert the legacy command name to its new name.
+            if name in self._legacy:
+                name = self._legacy[name]
+
+            found = None
             if name in self.commands:
-                cmd_ep = self.commands[name]
+                found = name
+            else:
+                candidates = _get_commands_by_partial_name(
+                    argv[:i], self.commands)
+                if len(candidates) == 1:
+                    found = candidates[0]
+            if found:
+                cmd_ep = self.commands[found]
                 if hasattr(cmd_ep, 'resolve'):
                     cmd_factory = cmd_ep.resolve()
                 else:
                     # NOTE(dhellmann): Some fake classes don't take
                     # require as an argument. Yay?
-                    arg_spec = inspect.getargspec(cmd_ep.load)
+                    arg_spec = utils.getargspec(cmd_ep.load)
                     if 'require' in arg_spec[0]:
                         cmd_factory = cmd_ep.load(require=False)
                     else:
                         cmd_factory = cmd_ep.load()
-                return (cmd_factory, name, search_args)
+                return (cmd_factory, return_name, search_args)
         else:
             raise ValueError('Unknown command %r' %
                              (argv,))
+
+    def _get_last_possible_command_index(self, argv):
+        """Returns the index after the last argument
+        in argv that can be a command word
+        """
+        for i, arg in enumerate(argv):
+            if arg.startswith('-'):
+                return i
+        return len(argv)
+
+    def add_command_group(self, group=None):
+        """Adds another group of command entrypoints"""
+        if group:
+            self.load_commands(group)
+
+    def get_command_groups(self):
+        """Returns a list of the loaded command groups"""
+        return self.group_list
+
+    def get_command_names(self, group=None):
+        """Returns a list of commands loaded for the specified group"""
+        group_list = []
+        if group is not None:
+            for ep in pkg_resources.iter_entry_points(group):
+                cmd_name = (
+                    ep.name.replace('_', ' ')
+                    if self.convert_underscores
+                    else ep.name
+                )
+                group_list.append(cmd_name)
+            return group_list
+        return list(self.commands.keys())
